#! Clarifai data upload

import importlib
import inspect
import os
import sys
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from multiprocessing import cpu_count
from typing import Iterator, Optional, Tuple, Union

from clarifai_grpc.grpc.api import resources_pb2, service_pb2, service_pb2_grpc
from clarifai_grpc.grpc.api.status import status_code_pb2
from tqdm import tqdm

from clarifai.client import create_stub
from clarifai.data_upload.datasets.base import Chunker
from clarifai.data_upload.datasets.image import (VisualClassificationDataset,
                                                 VisualDetectionDataset, VisualSegmentationDataset)
from clarifai.data_upload.datasets.text import TextClassificationDataset


def load_dataset(module_dir: Union[str, os.PathLike], split: str) -> Iterator:
  """
  Validate and import dataset module data generator.
  Args:
    `module_dir`: relative path to the module directory
    The directory must contain a `dataset.py` script and the data itself.
    `split`: "train" or "val"/"test" dataset split
  Module Directory Structure:
  ---------------------------
      <folder_name>/
      ├──__init__.py
      ├──<Your local dir dataset>/
      └──dataset.py
  dataset.py must implement a class named following the convention,
  <dataset_name>Dataset and this class must have a dataloader()
  generator method
  """
  sys.path.append(str(module_dir))

  if not os.path.exists(os.path.join(module_dir, "__init__.py")):
    with open(os.path.join(module_dir, "__init__.py"), "w"):
      pass

  import dataset  # dataset module

  # get main module class
  main_module_cls = None
  for name, obj in dataset.__dict__.items():
    if inspect.isclass(obj) and "Dataset" in name:
      main_module_cls = obj
    else:
      continue

  return main_module_cls(split).dataloader()


def load_zoo_dataset(name: str, split: str) -> Iterator:
  """
  Get dataset generator object from dataset zoo.
  Args:
    `name`: dataset module name in datasets/zoo/.
    `split`: "train" or "val"/"test" dataset split
  Returns:
    Data generator object
  """
  zoo_dataset = importlib.import_module(f"datasets.zoo.{name}")
  # get main module class
  main_module_cls = None
  for name, obj in zoo_dataset.__dict__.items():
    if inspect.isclass(obj) and "Dataset" in name:
      main_module_cls = obj
    else:
      continue

  return main_module_cls(split).dataloader()


class UploadConfig:

  def __init__(
      self,
      user_id: str,
      app_id: str,
      pat: str,
      dataset_id: str,
      task: str,
      from_module: Optional[Union[str, os.PathLike]] = None,
      from_zoo: Optional[str] = None,  # load dataset from zoo
      split: str = "train",  # train or test/val
      chunk_size: int = 16,
      portal: str = "clarifai"):
    """
    Initialize upload configs.
    Args:
      `user_id`: Clarifai user id.
      `app_id`: Clarifai app id.
      `pat`: Clarifai PAT(Personal Access Token).
      `dataset_id`: Clarifai dataset id (where data is to be uploaded).
      `task`: either of `visual_clf`, `visual_detection`, `visual_segmentation` or `text_clf`.
      `from_module`: Path to dataset module directory.
        Should be left as None if `from_zoo` is to be used.
      `from_zoo`: Name of dataset to upload from the zoo.
        The name must match the dataset module name excluding the file extension.
        Should be left as None if `from_module` is to be used.
      `split`: Dataset split to upload. Either of train or test/val
      `chunk_size`: size of chunks for parallel data upload.
    """
    self.USER_ID = user_id
    self.APP_ID = app_id
    self.PAT = pat
    self.dataset_id = dataset_id
    self.task = task
    self.module_dir = from_module
    self.zoo_dataset = from_zoo
    self.split = split
    self.chunk_size = chunk_size
    self.num_workers: int = cpu_count()
    self.__base: str = ""
    if portal == "dev":
      self.__base = "https://api-dev.clarifai.com"
    elif portal == "staging":
      self.__base = "https://api-staging.clarifai.com"
    else:  #prod
      self.__base = "https://api.clarifai.com"

    # Set auth vars as env variables
    os.environ["CLARIFAI_USER_ID"] = self.USER_ID
    os.environ["CLARIFAI_APP_ID"] = self.APP_ID
    os.environ["CLARIFAI_API_BASE"] = self.__base
    os.environ["CLARIFAI_PAT"] = self.PAT

    self.STUB: service_pb2_grpc.V2Stub = create_stub()
    self.metadata: Tuple = (('authorization', 'Key ' + self.PAT),)
    self.user_app_id = resources_pb2.UserAppIDSet(user_id=self.USER_ID, app_id=self.APP_ID)

  def _upload_inputs(self, inputs):
    """
    Upload inputs to clarifai platform dataset.
    Args:
      inputs: input protos
    """
    upload_count = 0
    retry_upload = []  # those that fail to upload are stored for retries

    for inp_proto in inputs:
      response = self.STUB.PostInputs(
          service_pb2.PostInputsRequest(user_app_id=self.user_app_id, inputs=[inp_proto]),)

      MESSAGE_DUPLICATE_ID = "Input has a duplicate ID."
      if response.status.code != status_code_pb2.SUCCESS:
        try:
          if response.inputs[0].status.details != MESSAGE_DUPLICATE_ID:
            retry_upload.append(inp_proto)
          print(f"Post inputs failed, status: {response.inputs[0].status.details}\n")
          continue
        except:
          print(f"Post inputs failed, status: {response.status.details}\n")
      else:
        upload_count += 1

    return retry_upload

  def upload_annotations(self, inputs):
    """
    Upload image annotations to clarifai detection dataset
    """
    upload_count = 0
    retry_upload = []  # those that fail to upload are stored for retries

    for annot_proto in inputs:
      response = self.STUB.PostAnnotations(
          service_pb2.PostAnnotationsRequest(
              user_app_id=self.user_app_id, annotations=[annot_proto]),)

      if response.status.code != status_code_pb2.SUCCESS:
        try:
          print(f"Post annotations failed, status:\n{response.annotations[0].status.details}\n")
          continue
        except:
          print(f"Post annotations failed, status:\n{response.status.details}\n")
          retry_upload.append(annot_proto)
      else:
        upload_count += 1

    return retry_upload

  def concurrent_inp_upload(self, inputs, chunks):
    """
    Upload images concurrently.
    """
    inp_threads = []
    retry_upload = []

    with ThreadPoolExecutor(max_workers=self.num_workers) as executor:
      for inp_batch in tqdm(inputs, total=chunks + 1, desc="uploading inputs..."):
        inp_threads.append(executor.submit(self._upload_inputs, inp_batch))
        time.sleep(0.1)

    for job in tqdm(
        as_completed(inp_threads), total=chunks + 1, desc="retry uploading failed protos..."):
      if job.result():
        retry_upload.extend(job.result())

    if len(
        list(retry_upload)) > 0:  ## TODO: use api_with_retries functionality via upload_inputs()
      _ = self._upload_inputs(retry_upload)

  def concurrent_annot_upload(self, inputs, chunks):
    """
    Uploads annotations concurrently.
    """
    annot_threads = []
    retry_annot_upload = []

    with ThreadPoolExecutor(max_workers=self.num_workers) as executor:
      for annot_batch in tqdm(inputs, total=chunks + 1, desc="uploading..."):
        annot_threads.append(executor.submit(self.upload_annotations, annot_batch))
        time.sleep(0.2)

    for job in tqdm(
        as_completed(annot_threads), total=chunks + 1, desc="retry uploading failed protos..."):
      if job.result():
        retry_annot_upload.extend(job.result())
    if len(retry_annot_upload) > 0:
      ## TODO: use api_with_retries functionality via upload_annotations()
      _ = self.upload_annotations(retry_annot_upload)

  def upload_to_clarifai(self):
    """
    Execute data upload.
    """
    datagen_object = None
    if self.module_dir is None and self.zoo_dataset is None:
      raise Exception("One of `from_module` and `from_zoo` must be \
      specified. Both can't be None or defined at the same time.")
    elif self.module_dir is not None and self.zoo_dataset is not None:
      raise Exception("Use either of `from_module` or `from_zoo` \
      but NOT both.")
    elif self.module_dir is not None:
      datagen_object = load_dataset(self.module_dir, self.split)
    else:
      datagen_object = load_zoo_dataset(self.zoo_dataset, self.split)

    if self.task == "text_clf":
      dataset_obj = TextClassificationDataset(datagen_object, self.dataset_id, self.split)
      text_protos = dataset_obj._get_input_protos()
      text_protos = dataset_obj._to_list(text_protos)

      # Upload text
      chunks = len(text_protos) // self.num_workers
      chunked_text_protos = Chunker(text_protos, self.chunk_size).chunk()

      self.concurrent_inp_upload(chunked_text_protos, chunks)

    elif self.task == "visual_detection":
      dataset_obj = VisualDetectionDataset(datagen_object, self.dataset_id, self.split)
      img_protos, annotation_protos = dataset_obj._get_input_protos()
      img_protos = dataset_obj._to_list(img_protos)

      # Upload images
      chunks = len(img_protos) // self.num_workers
      chunked_img_protos = Chunker(img_protos, self.chunk_size).chunk()

      self.concurrent_inp_upload(chunked_img_protos, chunks)

      # Upload annotations:
      print("Uploading annotations.......")
      annotation_protos = dataset_obj._to_list(annotation_protos)
      chunks_ = len(annotation_protos) // self.num_workers
      chunked_annot_protos = Chunker(annotation_protos, self.chunk_size).chunk()

      self.concurrent_annot_upload(chunked_annot_protos, chunks_)

    elif self.task == "visual_segmentation":
      dataset_obj = VisualSegmentationDataset(datagen_object, self.dataset_id, self.split)
      img_protos, mask_protos = dataset_obj._get_input_protos()
      img_protos = dataset_obj._to_list(img_protos)
      mask_protos = dataset_obj._to_list(mask_protos)

      # Upload images
      chunks = len(img_protos) // self.num_workers
      chunked_img_protos = Chunker(img_protos, self.chunk_size).chunk()

      #self.concurrent_inp_upload(chunked_img_protos, chunks)
      # Upload masks:
      print("Uploading masks.......")
      chunks_ = len(mask_protos) // self.num_workers
      chunked_mask_protos = Chunker(mask_protos, self.chunk_size).chunk()

      self.concurrent_annot_upload(chunked_mask_protos, chunks_)
    else:  # visual-classification & visual-captioning
      dataset_obj = VisualClassificationDataset(datagen_object, self.dataset_id, self.split)
      img_protos = dataset_obj._get_input_protos()
      img_protos = dataset_obj._to_list(img_protos)

      # Upload images
      chunks = len(img_protos) // self.num_workers
      chunked_img_protos = Chunker(img_protos, self.chunk_size).chunk()

      self.concurrent_inp_upload(chunked_img_protos, chunks)



from typing import Iterator, List

from clarifai_grpc.grpc.api import resources_pb2
from google.protobuf.struct_pb2 import Struct


class ClarifaiDataset:
  """
  Clarifai datasets base class.
  """

  def __init__(self, datagen_object: Iterator, dataset_id: str, split: str) -> None:
    self.datagen_object = datagen_object
    self.dataset_id = dataset_id
    self.split = split
    self._all_input_protos = []

  def __len__(self) -> int:
    """
    Get size of all input protos
    """
    return len(self._all_input_protos)

  def _to_list(self, input_protos: Iterator) -> List:
    """
    Parse protos iterator to list.
    """
    return list(input_protos)

  def create_input_protos(self, image_path: str, label: str, input_id: str, dataset_id: str,
                          metadata: Struct) -> resources_pb2.Input:
    """
    Create input protos for each image, label input pair.
    Args:
    	`image_path`: path to image.
    	`label`: image label
    	`input_id: unique input id
    	`dataset_id`: Clarifai dataset id
    	`metadata`: input metadata
    Returns:
    	An input proto representing a single row input
    """
    raise NotImplementedError()

  def _get_input_protos(self) -> Iterator:
    """
    Create input protos for each row of the dataframe.
    Returns:
    	A list of input protos
    """
    raise NotImplementedError()


class Chunker:
  """
  Split an input sequence into small chunks.
  """

  def __init__(self, seq: List, size: int) -> None:
    self.seq = seq
    self.size = size

  def chunk(self) -> List[List]:
    """
    Chunk input sequence.
    """
    return [self.seq[pos:pos + self.size] for pos in range(0, len(self.seq), self.size)]



import os
import tempfile
import zipfile
from io import BytesIO

import requests
from clarifai_grpc.grpc.api import resources_pb2
from PIL import ImageFile
from tqdm import tqdm


class DatasetExportReader:
  """
  Unpacks the zipfile from DatasetVersionExport
  - Downloads the temp archive onto disk
  - Reads DatasetVersionExports archive in memory without extracting all
  - Yield each api.Input object.
  """

  def __init__(self, archive_url=None, local_archive_path=None):

    self.input_count = 0
    self.temp_file = None

    assert archive_url or local_archive_path, "Must use one input."

    if archive_url:
      print('url: %s' % archive_url)
      self.temp_file = self._download_temp_archive(archive_url)
      self.archive = zipfile.ZipFile(self.temp_file)
    else:
      print("path: %s" % local_archive_path)
      self.archive = zipfile.ZipFile(local_archive_path)

    self.file_name_list = self.archive.namelist()
    assert "mimetype" in self.file_name_list, "Missing mimetype file in the dataset export archive."
    assert self.archive.read("mimetype") == b"application/x.clarifai-data+protobuf"
    self.file_name_list.remove("mimetype")

    print("Obtained file name list. %d entries." % len(self.file_name_list))
    self.split_dir = os.path.dirname(self.file_name_list[0]) if len(self.file_name_list) else ""

  def _download_temp_archive(self, archive_url, chunk_size=128):
    """
    Downloads the temp archive of InputBatches.
    """
    r = requests.get(archive_url, stream=True)
    temp_file = tempfile.TemporaryFile()
    for chunk in r.iter_content(chunk_size=chunk_size):
      temp_file.write(chunk)

    return temp_file

  def __len__(self):
    if not self.input_count:
      if self.file_name_list is not None:
        for filename in self.file_name_list:
          self.input_count += int(filename.split('_n')[-1])

    return self.input_count

  def __iter__(self):
    """
    Loops through all InputBatches in the DatasetVersionExport and yields every api.Input object
    """
    if self.file_name_list is not None:
      for filename in self.file_name_list:
        db = resources_pb2.InputBatch().FromString(self.archive.read(filename))
        for db_input in db.inputs:
          yield db_input
      print("DONE")

  def __enter__(self):
    return self

  def __exit__(self, *args):
    self.close()

  def close(self):
    print("closing file objects.")
    self.archive.close()
    if self.temp_file:
      self.temp_file.close()


class InputDownloader:
  """
  Takes an iterator or a list of api.Input instances as input,
  and has a method for downloading all inputs (currently only images) of that data.
  Has the ability of either writing to a new ZIP archive OR a filesystem directory.
  """

  def __init__(self, input_iterator):
    self.input_iterator = input_iterator
    self.num_inputs = 0
    self.split_prefix = None
    if isinstance(self.input_iterator, DatasetExportReader):
      self.split_prefix = self.input_iterator.split_dir

  def _save_image_to_archive(self, new_archive, hosted_url, name):
    """
    Use PIL ImageFile to return image parsed from the response bytestring (from requests) and append to zip file.
    """
    p = ImageFile.Parser()
    p.feed(requests.get(hosted_url).content)
    image = p.close()
    image_file = BytesIO()
    image.save(image_file, 'JPEG')
    new_archive.writestr(name, image_file.getvalue())

  def _write_image_archive(self, save_path, split):
    """
    Writes the image archive into prefix dir.
    """
    try:
      total = len(self.input_iterator)
    except TypeError:
      total = None
    with zipfile.ZipFile(save_path, "a") as new_archive:
      for input_ in tqdm(self.input_iterator, desc="Writing image archive", total=total):
        # checks for image
        hosted = input_.data.image.hosted
        if hosted.prefix:
          assert 'orig' in hosted.sizes
          hosted_url = f"{hosted.prefix}/orig/{hosted.suffix}"
          full_name = os.path.join(split, input_.id + ".jpg")

          self._save_image_to_archive(new_archive, hosted_url, full_name)
          self.num_inputs += 1

  def _check_output_archive(self, save_path):
    try:
      archive = zipfile.ZipFile(save_path, 'r')
    except zipfile.BadZipFile as e:
      raise e
    assert len(
        archive.namelist()) == self.num_inputs, "Archive has %d inputs | expecting %d inputs" % (
            len(archive.namelist()), self.num_inputs)

  def download_image_archive(self, save_path, split=None):
    """
    Downloads the archive from the URL into an archive of images in the directory format {split}/{image}.
    """
    self._write_image_archive(save_path, split=split or self.split_prefix)
    self._check_output_archive(save_path)


if __name__ == "__main__":
  import sys
  if len(sys.argv) < 2:
    print(f"usage: {sys.argv[0]} <archive-url> [<save-path>]")
    sys.exit(2)
  archive_url = sys.argv[1]
  save_path = sys.argv[2] if len(sys.argv) > 2 else "output.zip"

  with DatasetExportReader(archive_url=archive_url) as reader:
    InputDownloader(reader).download_image_archive(save_path=save_path)



#! Execute dataset upload using the `from_module` upload feature

from clarifai.data_upload.upload import UploadConfig

text_upload_obj = UploadConfig(
    user_id="",
    app_id="",
    pat="",
    dataset_id="",
    task="visual_clf",
    from_module="./examples/image_classification/cifar10",
    split="train",
    portal="clarifai"  #clarifai(prod), dev or staging
)
## change the task and from_module arguments in UploadConfig() to upload
## example food-101 dataset

if __name__ == "__main__":
  text_upload_obj.upload_to_clarifai()



from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2

from clarifai.client import V2Stub


def models_generator(stub: V2Stub,
                     user_id: str,
                     app_id: str,
                     page_size: int = 64,
                     only_in_app: bool = False):
  """
    Lists all the models accessible in an application given a userAppID user_id, app_id app.

    Args:
      stub: client stub.
      user_id: the user to list from.
      app_id: the app in the user_id account to list from.
      page_size: the pagination size to use while iterating.
      only_in_app: if the models returned should only be ones that have been created in the app

    Returns:
      models: a list of Model protos for all the models in the app.
    """
  userDataObject = resources_pb2.UserAppIDSet(user_id=user_id, app_id=app_id)

  model_success_status = {status_code_pb2.SUCCESS}

  page = 1
  while True:
    response = stub.ListModels(
        service_pb2.ListModelsRequest(user_app_id=userDataObject, page=page, per_page=page_size),)

    if response.status.code not in model_success_status:
      raise Exception("ListModels failed with response %r" % response)
    if len(response.models) == 0:
      break
    else:
      if only_in_app:
        for item in response.models:
          if item.app_id == app_id:
            yield item
      else:
        for item in response.models:
          yield item
    page += 1



import glob
import importlib


class ClarifaiModulePageManager(object):

  def __init__(self):
    # List all the available pages.
    page_files = sorted(glob.glob("pages/*.py"))
    self.page_names = [f.replace("pages/", "").replace(".py", "") for f in page_files]

  def get_page_from_query_params(self, qp):
    """
        Args:
          qp: the streamlit query params st.experimental_get_query_params()
        """
    # Get the page from query params or default to 1 from the url.
    page = qp.get("page", [None])[0]
    if page is None:
      page = self.page_names[0]
    # Check that the page number coming in is within the range of pages in the folder.
    if page not in self.page_names:
      raise Exception(
          "Page '%s' is not valid, there is no pages/%s.py file for this page. Valid page names are: %s"
          % (page, page, str(self.page_names)))

    return page

  def get_page_names(self):
    return self.page_names

  def render_page(self, page):
    # Since the page re-renders every time the selectbox changes, we'll always have the latest page out
    # of the query params.
    module_str = "pages.%s" % page
    # check if the page exists
    importlib.util.find_spec(module_str)
    if page is None:
      raise Exception("Page %s is was not found." % page)

    current_page = importlib.import_module(module_str)
    current_page.display()



#! ImageNet Classification dataset

import os

from clarifai.data_upload.datasets.features import VisualClassificationFeatures


class ImageNetDataset:
  """ImageNet Dataset."""

  def __init__(self, split: str = "train"):
    """
    Initialize dataset params.
    Args:
      data_dir: the local dataset directory.
      split: "train" or "test"
    """
    self.split = split
    self.data_dir = os.path.join(os.curdir, ".data")  # data storage directory
    self.label_map = dict()

  def dataloader(self):
    """
    Transform text data into clarifai proto compatible
    format for upload.
    Returns:
      VisualClassificationFeatures type generator.
    """
    #Creating label map
    with open(os.path.join(self.data_dir, "LOC_synset_mapping.txt")) as _file:
      for _id in _file:
        self.label_map[_id.split(' ')[0]] = [
            label.rstrip().lstrip() for label in _id[_id.find(' ') + 1:].split(',')
        ]

    for _folder in os.listdir(os.path.join(self.data_dir, self.split)):
      try:
        concept = self.label_map[_folder]  #concepts
      except:
        continue
      folder_path = os.path.join(self.data_dir, self.split) + "/" + _folder
      for _img in os.listdir(folder_path):
        if _img.lower().endswith(('.png', '.jpg', '.jpeg', '.tiff')):
          yield VisualClassificationFeatures(
              image_path=folder_path + "/" + _img, label=concept, id=None)



from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2

from clarifai.client import V2Stub


def modules_generator(stub: V2Stub, user_id: str, app_id: str, page_size: int = 64):
  """
  Lists all the modules in the given userAppID user_id, app_id app.

  Args:
    stub: client stub.
    user_id: the user to list from.
    app_id: the app in the user_id account to list from.
    page_size: the pagination size to use while iterating.

  Returns:
    imvs: a list of Module protos for all the modules in the app.
  """
  userDataObject = resources_pb2.UserAppIDSet(user_id=user_id, app_id=app_id)

  success_status = {status_code_pb2.SUCCESS}

  page = 1
  while True:
    response = stub.ListModules(
        service_pb2.ListModulesRequest(user_app_id=userDataObject, page=page, per_page=page_size),)

    if response.status.code not in success_status:
      raise Exception("ListModules failed with response %r" % response)
    for item in response.modules:
      yield item
    page += 1
    # if we don't get a full page back we know we're done.
    if len(response.modules) < page_size:
      break



#! COCO 2017 Image Segmentation dataset

import gc
import os
import zipfile
from functools import reduce
from glob import glob

import cv2
import numpy as np
import requests
from pycocotools import mask as maskUtils
from pycocotools.coco import COCO
from tqdm import tqdm

from ..features import VisualSegmentationFeatures


class COCOSegmentationDataset:
  """COCO 2017 Image Segmentation Dataset."""

  def __init__(self, split: str = "train"):
    """
    Initialize coco dataset.
    Args:
      filenames: the coco zip filenames: List[str] to be downloaded if download=True,
      data_dir: the local coco dataset directory
      split: "train" or "val"
    """
    self.filenames = {
        "train": "train2017.zip",
        "val": "val2017.zip",
        "annotations": "annotations_trainval2017.zip"
    }
    self.split = split
    self.url = "http://images.cocodataset.org/zips/"  # coco base image-zip url
    self.data_dir = os.path.join(os.curdir, ".data")  # data storage dir
    self.extracted_coco_dirs = {"train": None, "val": None, "annotations": None}

  def coco_download(self, save_dir):
    """Download coco dataset."""
    if not os.path.exists(save_dir):
      os.mkdir(save_dir)

    #check if train, val and annotation dirs exist
    #so that the coco2017 data isn't downloaded
    for key, filename in self.filenames.items():
      if os.path.exists(glob(f"{save_dir}/{key}*")[0]):
        print("dataset already downloded and extracted")
        continue

      print("-" * 80)
      print(f"Downloading {filename}")
      print("-" * 80)

      if "annotations" in filename:
        self.url = "http://images.cocodataset.org/annotations/"

      response = requests.get(self.url + filename, stream=True)
      response.raise_for_status()
      with open(os.path.join(save_dir, filename), "wb") as _file:
        for chunk in tqdm(response.iter_content(chunk_size=5124000)):
          if chunk:
            _file.write(chunk)
      print("Coco data download complete...")

      #extract files
      zf = zipfile.ZipFile(os.path.join(save_dir, filename))
      print(f" Extracting {filename} file")
      zf.extractall(path=save_dir)
      # Delete coco zip
      print(f" Deleting {filename}")
      os.remove(path=os.path.join(save_dir, filename))

  def dataloader(self):
    """
    Transform coco data into clarifai proto compatible format for upload.
    Returns:
      VisualSegmentationFeatures type generator.
    """
    if isinstance(self.filenames, list) and len(self.filenames) == 3:
      self.coco_download(self.data_dir)
      self.extracted_coco_dirs["train"] = [os.path.join(self.data_dir, i) \
      for i in os.listdir(self.data_dir) if "train" in i][0]
      self.extracted_coco_dirs["val"] = [os.path.join(self.data_dir, i) \
      for i in os.listdir(self.data_dir) if "val" in i][0]

      self.extracted_coco_dirs["annotations"] = [os.path.join(self.data_dir, i) \
      for i in os.listdir(self.data_dir) if "annotations" in i][0]
    else:
      raise Exception(f"`filenames` must be a list of atleast 3 coco zip file names; \
      train, val and annotations. Found {len(self.filenames)} items instead.")

    annot_file = glob(self.extracted_coco_dirs["annotations"] + "/" +\
     f"instances_{self.split}*")[0]
    coco = COCO(annot_file)
    categories = coco.loadCats(coco.getCatIds())
    cat_id_map = {category["id"]: category["name"] for category in categories}
    cat_img_ids = {}
    for cat_id in list(cat_id_map.keys()):
      cat_img_ids[cat_id] = coco.getImgIds(catIds=[cat_id])

    img_ids = []
    for i in list(cat_img_ids.values()):
      img_ids.extend(i)

    #get annotations for each image id
    for _id in img_ids:
      annots = []  # polygons
      class_names = []
      labels = [i for i in list(filter(lambda x: _id in cat_img_ids[x], cat_img_ids))]
      image_path = glob(self.extracted_coco_dirs[self.split]+"/"+\
      f"{str(_id).zfill(12)}*")[0]

      image_height, image_width = cv2.imread(image_path).shape[:2]
      for cat_id in labels:
        annot_ids = coco.getAnnIds(imgIds=_id, catIds=[cat_id])
        if len(annot_ids) > 0:
          img_annotations = coco.loadAnns(annot_ids)
          for ann in img_annotations:
            class_names.append(cat_id_map[cat_id])
            # get polygons
            if type(ann['segmentation']) == list:
              for seg in ann['segmentation']:
                poly = np.array(seg).reshape((int(len(seg) / 2), 2))
                poly[:, 0], poly[:, 1] = poly[:, 0] / image_width, poly[:, 1] / image_height
                annots.append(poly.tolist())  #[[x=col, y=row],...]
            else:  # seg: {"counts":[...]}
              if type(ann['segmentation']['counts']) == list:
                rle = maskUtils.frPyObjects([ann['segmentation']], image_height, image_width)
              else:
                rle = ann['segmentation']
              mask = maskUtils.decode(rle)  #binary mask
              #convert mask to polygons and add to annots
              contours, _ = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
              polygons = []
              for cont in contours:
                if cont.size >= 6:
                  polygons.append(cont.astype(float).flatten().tolist())
              # store polygons in (x,y) pairs
              polygons_flattened = reduce(lambda x, y: x + y, polygons)
              del polygons
              del contours
              del mask
              gc.collect()

              polygons = np.array(polygons_flattened).reshape((int(len(polygons_flattened) / 2),
                                                               2))
              polygons[:, 0] = polygons[:, 0] / image_width
              polygons[:, 1] = polygons[:, 1] / image_height

              annots.append(polygons.tolist())  #[[x=col, y=row],...,[x=col, y=row]]
        else:  # if no annotations for given image_id-cat_id pair
          continue
      assert len(class_names) == len(annots), f"Num classes must match num annotations\
      for a single image. Found {len(class_names)} classes and {len(annots)} polygons."

      yield VisualSegmentationFeatures(image_path, class_names, annots, _id)



#! Cifar10 Dataset

import csv
import os

from clarifai.data_upload.datasets.features import VisualClassificationFeatures


class Cifar10Dataset:
  """Cifar10 Dataset."""

  def __init__(self, split: str = "train"):
    """
    Initialize dataset params.
    Args:
      data_dir: the local dataset directory.
      split: "train" or "test"
    """
    self.split = split
    self.data_dirs = {
        "train": os.path.join(os.path.dirname(__file__), "cifar_small_train.csv"),
        "test": os.path.join(os.path.dirname(__file__), "cifar_small_test.csv")
    }

  def dataloader(self):
    """
    Transform text data into clarifai proto compatible
    format for upload.
    Returns:
      TextFeatures type generator.
    """
    ## Your preprocessing code here
    with open(self.data_dirs[self.split]) as _file:
      reader = csv.reader(_file)
      next(reader, None)  # skip header
      for review in reader:
        yield VisualClassificationFeatures(
            image_path='examples/image_classification/cifar10/' + review[0],
            label=review[1],
            id=None)



from clarifai_grpc.grpc.api import resources_pb2

from clarifai.client import V2Stub
from clarifai.listing.concepts import concepts_generator
from clarifai.listing.datasets import datasets_generator
from clarifai.listing.inputs import dataset_inputs_generator, inputs_generator
from clarifai.listing.installed_module_versions import installed_module_versions_generator
from clarifai.listing.models import models_generator
from clarifai.listing.module_versions import module_versions_generator
from clarifai.listing.modules import modules_generator


class ClarifaiResourceLister(object):

  def __init__(self, stub: V2Stub, user_id: str, app_id: str, page_size: int = 16):
    """
        Helper class for common listing of resources in an Clarifai App.

        Args:
          stub: grpc client V2Strub for our API.
          user_id: the user to list from.
          app_id: the app in the user_id account to list from.
          page_size: the pagination size to use while iterating.
        """
    self.stub = stub
    self.user_id = user_id
    self.app_id = app_id
    self.user_app_id_proto = resources_pb2.UserAppIDSet(user_id=user_id, app_id=app_id)
    self.default_page_size = page_size

  def list_all_models(self, page_size: int = None, only_in_app: bool = False):
    """
        This lists all the Models accessible in app. Not recommended for large apps.

        Params:
          page_size: how many elements per page to fetch
          only_in_app: if only models created in the app should be returned

        Returns:
          inputs: a list of Model protos for all the inputs in the app.
        """
    return [item for item in self.models_generator(page_size, only_in_app)]

  def models_generator(self, page_size: int = None, only_in_app: bool = False):
    """
        This lists all the models in an app. Not recommended for large apps.

        Params:
          page_size: how many elements per page to fetch
          only_in_app: if only models created in the app should be returned

        Returns:
          gen: a generator that yields a single Model proto at a time.
        """
    page_size = self.default_page_size if page_size is None else page_size
    return models_generator(self.stub, self.user_id, self.app_id, page_size, only_in_app)

  def list_all_datasets(self, page_size: int = None):
    """
        This lists all Datasets in app. Not recommended for large apps.

        Params:
          page_size: how many elements per page to fetch

        Returns:
          inputs: a list of Dataset protos for all the inputs in the app.
        """
    return [item for item in self.datasets_generator(page_size)]

  def datasets_generator(self, page_size: int = None):
    """
        This lists all the Datasets in an app. Not recommended for large apps.

        Params:
          page_size: how many elements per page to fetch

        Returns:
          gen: a generator that yields a single Dataset proto at a time.
        """
    page_size = self.default_page_size if page_size is None else page_size
    return datasets_generator(self.stub, self.user_id, self.app_id, page_size)

  def list_all_concepts(self, page_size: int = None):
    """
        This lists all the concepts in an app. Not recommended for large apps.

        Returns:
          concepts: a list of Concept protos for all the concepts in the app.
        """
    return [item for item in self.concepts_generator(page_size)]

  def concepts_generator(self, page_size: int = None):
    """
        This lists all the concepts in an app. Not recommended for large apps.

        Returns:
          gen: a generator that yields a single Concept proto at a time.
        """
    page_size = self.default_page_size if page_size is None else page_size
    return concepts_generator(self.stub, self.user_id, self.app_id, page_size)

  def list_all_inputs(self, page_size: int = None):
    """
        This lists all the inputs in an app. Not recommended for large apps.

        Returns:
          inputs: a list of Input protos for all the inputs in the app.
        """
    return [item for item in self.inputs_generator(page_size)]

  def inputs_generator(self, page_size: int = None):
    """
        This lists all the inputs in an app. Not recommended for large apps.

        Returns:
          gen: a generator that yields a single Input proto at a time.
        """
    page_size = self.default_page_size if page_size is None else page_size
    return inputs_generator(self.stub, self.user_id, self.app_id, page_size)

  def list_all_dataset_inputs(self, page_size: int = None, dataset_id: str = None):
    """
        This lists all the inputs in a dataset. Not recommended for large datasets.

        Returns:
          inputs: a list of Input protos for all the inputs in the app.
        """
    return [item for item in self.dataset_inputs_generator(page_size, dataset_id)]

  def dataset_inputs_generator(self, page_size: int = None, dataset_id: str = None):
    """
        This lists all the inputs in a dataset. Not recommended for large datasets.

        Returns:
          gen: a generator that yields a single Input proto at a time.
        """
    page_size = self.default_page_size if page_size is None else page_size
    return dataset_inputs_generator(
        stub=self.stub,
        user_id=self.user_id,
        app_id=self.app_id,
        page_size=page_size,
        dataset_id=dataset_id)

  def list_all_installed_module_versions(self, page_size: int = None):
    """
    This lists all the installed_module_versions in an app. Not recommended for large apps.

    Returns:
      installed_module_versions: a list of InstalledModuleVersion protos for all the installed_module_versions in the app.
    """
    return [item for item in self.installed_module_versions_generator(page_size)]

  def installed_module_versions_generator(self, page_size: int = None):
    """
    This lists all the installed_module_versions in an app. Not recommended for large apps.

    Returns:
      gen: a generator that yields a single InstalledModuleVersion proto at a time.
    """
    page_size = self.default_page_size if page_size is None else page_size
    return installed_module_versions_generator(self.stub, self.user_id, self.app_id, page_size)

  def list_all_modules(self, page_size: int = None):
    """
    This lists all the modules in an app. Not recommended for large apps.

    Returns:
      modules: a list of Module protos for all the modules in the app.
    """
    return [item for item in self.module_generator(page_size)]

  def module_generator(self, page_size: int = None):
    """
    This lists all the module in an app. Not recommended for large apps.

    Returns:
      gen: a generator that yields a single Module proto at a time.
    """
    page_size = self.default_page_size if page_size is None else page_size
    return modules_generator(self.stub, self.user_id, self.app_id, page_size)

  def list_all_module_versions(self, module_id: str, page_size: int = None):
    """
    This lists all the module_versions in an app. Not recommended for large apps.

    Returns:
      module_versions: a list of ModuleVersion protos for all the module_versions in the app.
    """
    return [item for item in self.module_versions_generator(module_id, page_size)]

  def module_versions_generator(self, module_id: str, page_size: int = None):
    """
    This lists all the module_versions in an app. Not recommended for large apps.

    Returns:
      gen: a generator that yields a single ModuleVersion proto at a time.
    """
    page_size = self.default_page_size if page_size is None else page_size
    return module_versions_generator(self.stub, self.user_id, self.app_id, module_id, page_size)



#! COCO 2017 image captioning dataset

import os
import zipfile
from glob import glob

import requests
from pycocotools.coco import COCO
from tqdm import tqdm

from ..features import VisualClassificationFeatures


class COCOCaptionsDataset:
  """COCO 2017 Image Captioning Dataset."""

  def __init__(self, split: str = "train"):
    """
    Initialize coco dataset.
    Args:
      filenames: the coco zip filenames: List[str] to be downloaded if download=True,
      data_dir: the local coco dataset directory.
      split: "train" or "val"
    """
    self.filenames = {
        "train": "train2017.zip",
        "val": "val2017.zip",
        "annotations": "annotations_trainval2017.zip"
    }
    self.split = split
    self.url = "http://images.cocodataset.org/zips/"  # coco base image-zip url
    self.data_dir = os.path.join(os.curdir, ".data")  # data storage directory
    self.extracted_coco_dirs = {"train": None, "val": None, "annotations": None}

  def coco_download(self, save_dir):
    """Download coco dataset."""
    if not os.path.exists(save_dir):
      os.mkdir(save_dir)

    #check if train, val and annotation dirs exist
    #so that the coco2017 data isn't downloaded
    for key, filename in self.filenames.items():
      if os.path.exists(glob(f"{save_dir}/{key}*")[0]):
        print("Dataset already downloded and extracted")
        continue

      print("-" * 80)
      print(f"Downloading {filename}")
      print("-" * 80)

      if "annotations" in filename:
        self.url = "http://images.cocodataset.org/annotations/"

      response = requests.get(self.url + filename, stream=True)
      response.raise_for_status()
      with open(os.path.join(save_dir, filename), "wb") as _file:
        for chunk in tqdm(response.iter_content(chunk_size=5124000)):
          if chunk:
            _file.write(chunk)
      print("Data download complete...")

      #extract files
      zf = zipfile.ZipFile(os.path.join(save_dir, filename))
      print(f" Extracting {filename} file")
      zf.extractall(path=save_dir)
      # Delete coco zip
      print(f" Deleting {filename}")
      os.remove(path=os.path.join(save_dir, filename))

  def dataloader(self):
    """
    Transform coco image captioning data into clarifai proto compatible
    format for upload.
    Returns:
      VisualClassificationFeatures type generator.
    """
    if isinstance(self.filenames, list) and len(self.filenames) == 3:  #train, val, annotations
      self.coco_download(self.data_dir)
      self.extracted_coco_dirs["train"] = [os.path.join(self.data_dir, i) \
      for i in os.listdir(self.data_dir) if "train" in i][0]
      self.extracted_coco_dirs["val"] = [os.path.join(self.data_dir, i) \
      for i in os.listdir(self.data_dir) if "val" in i][0]

      self.extracted_coco_dirs["annotations"] = [os.path.join(self.data_dir, i) \
      for i in os.listdir(self.data_dir) if "annotations" in i][0]
    else:
      raise Exception(f"`filenames` must be a list of atleast 3 coco zip file names; \
      train, val and annotations. Found {len(self.filenames)} items instead.")

    annot_file = glob(self.extracted_coco_dirs["annotations"] + "/" + f"captions_{self.split}*")[0]
    coco = COCO(annot_file)
    annot_ids = coco.getAnnIds()
    annotations = coco.loadAnns(annot_ids)
    for annot in annotations:
      image_path = glob(self.extracted_coco_dirs[self.split]+"/"+\
      f"{str(annot['image_id']).zfill(12)}*")[0]
      # image_captioning and image classification datasets have the same
      # image-label input feature formats
      yield VisualClassificationFeatures(image_path, annot["caption"], annot["image_id"])
